#ifndef Matrix_h
#define Matrix_h

#include <stddef.h>

/**
 * M×N行列クラス。
 *
 * @param M 行の数。
 * @param N 列の数。
 * @param T 要素の型。
 */
template <size_t M, size_t N, class T = float>
class Matrix
{
  /**
   * 要素。
   */
  T m_elm[M * N] ;

public :
  /**
   * m_elm[] を初期化しないコンストラクタ。
   */
  Matrix() = default ;

  /**
   * コピーコンストラクタ。
   *
   * @param U       コピー元行列の要素型。
   * @param another コピー元行列
   */
  template <class U>
  Matrix(const Matrix<M, N, U>& another)
  {
    T *p = m_elm ;
    const U *q = (const U*)another ;
    for (T *pEnd = p + M*N ; p < pEnd ; ) *p++ = (T)*q++ ;
  }

  /**
   * 配列で値を指定して初期化。
   */
  Matrix(const T (&mat)[M][N])
  {
    for (size_t m = 0 ; m < M ; ++m) {
      for (size_t n = 0 ; n < N ; ++n) {
        m_elm[m + M * n] = mat[m][n] ;
      }
    }
  }

  /**
   * 要素の取得／設定。
   *
   * @param  m 行
   * @param  n 列
   * @return   m行n列の要素の値
   */
  T& at(size_t m, size_t n) { return m_elm[m + M * n] ; }
  const T& at(size_t m, size_t n) const { return m_elm[m + M * n] ; }

  /**
   * 要素配列へのポインタを取得。
   *
   * @return 要素(0,0)へのポインタ。
   */
  explicit operator const T *() const { return &m_elm[0]; }
  explicit operator T *() { return &m_elm[0]; }

public :
  /**
   * ゼロ行列。
   */
  static const Matrix O ;
} ;

template <size_t M, size_t N, class T>
const Matrix<M, N, T> Matrix<M, N, T>::O = {{0}};

// 4x4行列の場合、Matrix4<型名>と書けるようにする
template <class T = float>
using Matrix4 = Matrix<4, 4, T> ;

// さらに、4x4のfloat行列の場合、4x4のdouble配列の場合の別名を作る
typedef Matrix4<float> Matrix4f ;
typedef Matrix4<double> Matrix4d ;

#endif
