#ifndef gl__Matrix_h
#define gl__Matrix_h

#include "../noncopyable.h"
#include "../glall.h"
#include "../Matrix.h"
#include <stdexcept>

namespace gl
{
  /**
   * 一時的に行列スタックを変更するためのクラス。
   *
   * @param SetterConst glMatrixMode()に指定する定数。
   * @param GetterConst glGet*()に指定する定数。
   */
  template <GLenum SetterConst, GLenum GetterConst>
  class Matrix : noncopyable<Matrix<SetterConst, GetterConst>>
  {
    // gl::Contextクラスでのみインスタンス化可能。
    friend class Context ;

  private :
    /**
     * 現在値をスタックに保存。
     */
    Matrix()
    {
      // matrix modeを変更
      GLint mode ;
      glGetIntegerv(GL_MATRIX_MODE, &mode) ;
      if (mode != SetterConst) glMatrixMode(SetterConst) ;

      // 現在値を退避
      glPushMatrix() ;
      if (GL_STACK_OVERFLOW == glGetError()) {
        throw std::overflow_error("matrix stack overflow.") ;
      }

      // matri modeを元に戻す
      if (mode != SetterConst) glMatrixMode(mode) ;
    }

  public :
    /**
     * ムーブコンストラクタ。
     */
    Matrix(Matrix &&status) : m_moved(status.m_moved)
    {
      status.m_moved = true ;
    }

    /**
     * デストラクタ。
     */
    ~Matrix()
    {
      // 移動済みでない場合は初期値に戻す
      if (!m_moved) {
        // matrix modeを変更
        GLint mode ;
        glGetIntegerv(GL_MATRIX_MODE, &mode) ;
        if (mode != SetterConst) glMatrixMode(SetterConst) ;

        // 退避した値に戻す
        glPopMatrix() ;

        // matrix modeを元に戻す
        if (mode != SetterConst) glMatrixMode(mode) ;
      }
    }

    /**
     * 行列値を設定する。
     *
     * @param  enabled 新しい行列値。
     * @return         オブジェクト自身。
     */
    Matrix &operator = (const Matrix4f &mat)
    {
      // matrix modeを変更
      GLint mode ;
      glGetIntegerv(GL_MATRIX_MODE, &mode) ;
      if (mode != SetterConst) glMatrixMode(SetterConst) ;

      // 行列値の設定
      glLoadMatrixf((const float *)mat) ;

      // matrix modeを元に戻す
      if (mode != SetterConst) glMatrixMode(mode) ;

      return *this ;
    }

    /**
     * 行列値を取得。
     *
     * @return 現在の行列値。
     */
    operator Matrix4f() const
    {
      Matrix4f mat ;
      glGetFloatv(GetterConst, (float *)mat) ;

      return mat ;
    }

    /**
     * 行列を乗算する。
     */
    Matrix &operator *= (const Matrix4f &mat)
    {
      // matrix modeを変更
      GLint mode ;
      glGetIntegerv(GL_MATRIX_MODE, &mode) ;
      if (mode != SetterConst) glMatrixMode(SetterConst) ;

      // 乗算
      glMultMatrixf((const float *)mat) ;

      // matrix modeを元に戻す
      if (mode != SetterConst) glMatrixMode(mode) ;

      return *this ;
    }
    
  private :
    bool m_moved ;
  } ;
}

#endif 
